// SDL-Interface-Funktionen fr C / C++
//
// Klaus Kusche, 2012, vereinheitlichte Version fr alle Beispiele

#include <stdio.h>

#include "sdlinterf.h"
#include "SDL/SDL.h"

// ein Wert kleiner als der kleinste Koordinatenwert
#define MINVAL -1
// ein Wert grer als der grte Koordinatenwert
#define MAXVAL 9999

// kleinste / grte Koordinaten des rechteckigen Bildausschnittes,
// der vom nchsten sdlUpdate aktualisiert werden muss
// (ursprnglich: gar keiner).
static int minX = MAXVAL, maxX = MINVAL, minY = MAXVAL, maxY = MINVAL;

static SDL_Surface *myScreen;

// zyklische Farbtabelle mit 192 Farben fr Fraktale
struct colors {
  int r, g, b;
} colormap[192];

void sdlInit(void)
{
  int i;
  
  if (SDL_Init(SDL_INIT_VIDEO) == -1) { 
    fprintf(stderr, "Could not initialize SDL: %s.\n", SDL_GetError());
    exit(EXIT_FAILURE);
  }
  atexit(SDL_Quit);
  myScreen = SDL_SetVideoMode(SDL_X_SIZE, SDL_Y_SIZE, 0, SDL_SWSURFACE|SDL_ANYFORMAT);
  if (myScreen == NULL) {
    fprintf(stderr, "Could not set video mode: %s\n", SDL_GetError());
    exit(EXIT_FAILURE);
  }
  SDL_WM_SetCaption("SDL Graphics", "SDL Graphics");

  for (i = 0; i < 64; ++i) {
    colormap[i].r = 0;
    colormap[i].g = 4 * i;
    colormap[i].b = 255 - 4 * i;
    colormap[64 + i].r = 4 * i;
    colormap[64 + i].g = 255 - 4 * i;
    colormap[64 + i].b = 0;
    colormap[128 + i].r = 255 - 4 * i;
    colormap[128 + i].g = 0;
    colormap[128 + i].b = 4 * i;
  }
}

void sdlExit(void)
{
  SDL_Quit();
}

void sdlSetBlack(void)
{
  Uint32 pixel;

  pixel = SDL_MapRGB(myScreen->format, 0, 0, 0); 
  
  if (SDL_FillRect(myScreen, NULL, pixel) != 0) {
    fprintf(stderr, "Could not set screen to black: %s\n", SDL_GetError());
    exit(EXIT_FAILURE);
  }

  // alles ist modifiziert
  minX = 0;
  minY = 0;
  maxX = SDL_X_SIZE - 1;
  maxY = SDL_Y_SIZE - 1;
}

static inline void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
  int bpp = surface->format->BytesPerPixel;
  /* p = die Adresse im Bildspeicher, an die das Pixel gehrt */
  Uint8 *p = (Uint8 *)(surface->pixels) + y * surface->pitch + x * bpp;

  switch(bpp) {
  case 1:
    *p = pixel;
    break;
  case 2:
    *(Uint16 *)p = pixel;
    break;
  case 4:
    *(Uint32 *)p = pixel;
    break;
  default:
    fprintf(stderr, "Unknown screen depth: %d\n", bpp);
    exit(EXIT_FAILURE);
  }
}

void sdlDrawPoint(int x, int y, int r, int g, int b)
{
  if ((x < 0) || (x >= SDL_X_SIZE) ||
      (y < 0) || (y >= SDL_Y_SIZE)) {
    fprintf(stderr, "Point %d/%d is offscreen\n", x, y);
    exit(EXIT_FAILURE);
  }

  if (SDL_MUSTLOCK(myScreen)) {
    if (SDL_LockSurface(myScreen) < 0) {
      fprintf(stderr, "Could not lock screen: %s\n", SDL_GetError());
      exit(EXIT_FAILURE);
    }
  }

  putpixel(myScreen, x, y, SDL_MapRGB(myScreen->format, r, g, b));

  if (SDL_MUSTLOCK(myScreen)) {
    SDL_UnlockSurface(myScreen);
  }

  // ev. modifizierten Bereich vergrern
  if (x < minX) minX = x;
  if (y < minY) minY = y;
  if (x > maxX) maxX = x;
  if (y > maxY) maxY = y;
}

void sdlDrawCyclicPoint(int x, int y, int color)
{
  struct colors *cPtr;

  if (color < 0) {
    sdlDrawPoint(x, y, 0, 0, 0);
  } else {
    cPtr = &(colormap[color % 192]);
    sdlDrawPoint(x, y, cPtr->r, cPtr->g, cPtr->b);    
  }
}

void sdlDrawRect(int centerX, int centerY, int extX, int extY,
                 int r, int g, int b)
{
  Uint32 pixel;
  SDL_Rect rect;

  rect.x = centerX - extX;
  rect.y = centerY - extY;
  rect.w = extX + 1 + extX;
  rect.h = extY + 1 + extY;

  if ((rect.x < 0) || (rect.y < 0) ||
      (rect.w < 1) || (rect.h < 1)  ||
      (rect.x + rect.w > SDL_X_SIZE) || (rect.y + rect.h > SDL_Y_SIZE)) { 
    fprintf(stderr, "Rect (%d/%d +- %d/%d) offscreen.\n",
            centerX, centerY, extX, extY);
    exit(1);
  }
  
  pixel = SDL_MapRGB(myScreen->format, r, g, b); 
  
  if (SDL_FillRect(myScreen, &rect, pixel) != 0) {
    fprintf(stderr, "Could not fill rectangle: %s\n", SDL_GetError());
    exit(1);
  }

  // ev. modifizierten Bereich vergrern
  if (rect.x < minX) minX = rect.x;
  if (rect.y < minY) minY = rect.y;
  if ((rect.x + rect.w - 1) > maxX) maxX = rect.x + rect.w - 1;
  if ((rect.y + rect.h - 1) > maxY) maxY = rect.y + rect.h - 1;
}

void sdlDrawLine(int x1, int y1, int x2, int y2, int r, int g, int b)
{
  Uint32 pixel;
  int exchanged;
  int t;
  int deltaX, deltaY, stepY, error;
  int x, y;

  if ((x1 < 0) || (x1 >= SDL_X_SIZE) ||
      (y1 < 0) || (y1 >= SDL_Y_SIZE) ||
      (x2 < 0) || (x2 >= SDL_X_SIZE) ||
      (y2 < 0) || (y2 >= SDL_Y_SIZE)) {
    fprintf(stderr, "Line %d/%d  - %d/%d is offscreen\n",
            x1, y1, x2, y2);
    exit(EXIT_FAILURE);
  }

  // ev. modifizierten Bereich vergrern 
  // (vorher, weil x1/x2/y1/y2 unten verndert werden)
  if (x1 <= x2) {
    if (x1 < minX) minX = x1;
    if (x2 > maxX) maxX = x2;
  } else {
    if (x2 < minX) minX = x2;
    if (x1 > maxX) maxX = x1;
  }
  if (y1 <= y2) {
    if (y1 < minY) minY = y1;
    if (y2 > maxY) maxY = y2;
  } else {
    if (y2 < minY) minY = y2;
    if (y1 > maxY) maxY = y1;
  }

  pixel = SDL_MapRGB(myScreen->format, r, g, b); 

  if (SDL_MUSTLOCK(myScreen)) {
    if (SDL_LockSurface(myScreen) < 0) {
      fprintf(stderr, "Could not lock screen: %s\n", SDL_GetError());
      exit(EXIT_FAILURE);
    }
  }

  exchanged = (abs(y2 - y1) > abs(x2 - x1));
  if (exchanged) {
    t = x1; x1 = y1; y1 = t;
    t = x2; x2 = y2; y2 = t;
  }
  if (x1 > x2) {
    t = x1; x1 = x2; x2 = t;
    t = y1; y1 = y2; y2 = t;
  }
  deltaX = x2 - x1;
  deltaY = abs(y2 - y1);
  error = deltaX / 2;
  y = y1;
  if (y1 < y2) stepY = 1;
  else stepY = -1;
  for (x = x1; x <= x2; ++x) {
     if (exchanged) putpixel(myScreen, y, x, pixel);
     else putpixel(myScreen, x, y, pixel);
     error = error - deltaY;
     if (error < 0) {
       y = y + stepY;
       error = error + deltaX;
     }
  }

  if (SDL_MUSTLOCK(myScreen)) {
    SDL_UnlockSurface(myScreen);
  }
}

static void drawCircLine(int centerX, int centerY, int x1, int x2, int y,
                         Uint32 pixel, int part)
{
  int i, mY, pY;

  if (x1 < 0) x1 = 0;
  if (y < 0) y = 0;
  mY = centerY - y;
  pY = centerY + y;
  if (part & SDL_CIRC_FILLED) {
    x2 = 0;
  } else if (x2 > x1) {
    x2 = x1;
  }

  for (i = x1; i >= x2; --i) {
    if (part & SDL_CIRC_UPPER_LEFT)
      putpixel(myScreen, centerX - i, mY, pixel);
    if (part & SDL_CIRC_UPPER_RIGHT)
      putpixel(myScreen, centerX + i, mY, pixel);
    if (part & SDL_CIRC_LOWER_LEFT)
      putpixel(myScreen, centerX - i, pY, pixel);
    if (part & SDL_CIRC_LOWER_RIGHT)
      putpixel(myScreen, centerX + i, pY, pixel);
  }
}

void sdlDrawCirc(int centerX, int centerY, int radX, int radY,
                 int r, int g, int b)
{
  sdlDrawCircPart(centerX, centerY, radX, radY, r, g, b,
                  SDL_CIRC_ALL | SDL_CIRC_FILLED );
}

void sdlDrawCircPart(int centerX, int centerY, int radX, int radY,
                     int r, int g, int b, int part)
{
  Uint32 pixel;
  int x, y, oldX, lastX = 0, lastY = 0;
  int chgX, chgY;
  int err;
  int sqrX, sqrY;
  int stopX, stopY;
  int leftX, upperY, rightX, lowerY;

  leftX =  centerX - ((part & SDL_CIRC_LEFT) ? radX : 0);
  upperY = centerY - ((part & SDL_CIRC_UPPER) ? radY : 0);
  rightX = centerX + ((part & SDL_CIRC_RIGHT) ? radX : 0);
  lowerY = centerY + ((part & SDL_CIRC_LOWER) ? radY : 0);
  if ((radX < 0) || (radY < 0) ||
      (leftX < 0) || (upperY < 0) ||
      (rightX >= SDL_X_SIZE) || (lowerY >= SDL_Y_SIZE)) { 
    fprintf(stderr, "Circ %d/%d +- %d/%d is offscreen\n",
            centerX, centerY, radX, radY);
    exit(EXIT_FAILURE);
  }

  pixel = SDL_MapRGB(myScreen->format, r, g, b); 

  if (SDL_MUSTLOCK(myScreen)) {
    if (SDL_LockSurface(myScreen) < 0) {
      fprintf(stderr, "Could not lock screen: %s\n", SDL_GetError());
      exit(EXIT_FAILURE);
    }
  }

  // if both rad's are zero, the algorithm fails badly
  if ((radX == 0) && (radY == 0)) {
    putpixel(myScreen, centerX, centerY, pixel);
  } else {
    sqrX = 2 * radX * radX;
    sqrY = 2 * radY * radY;
    
    x = radX;
    y = 0;
    chgX = radY * radY * (1 - 2 * radX);
    chgY = radX * radX;
    err = 0;
    stopX = sqrY * radX;
    stopY = 0;

    while (stopX >= stopY) {
      drawCircLine(centerX, centerY, x, x, y, pixel, part);
      lastX = x; lastY = y;
      ++y;
      stopY += sqrX;
      err += chgY;
      chgY += sqrX;
      if (2 * err + chgX > 0) {
        --x;
        stopX -= sqrY;
        err += chgX;
        chgX += sqrY;
      }
    }

    x = 0;
    oldX = 0;
    y = radY;
    chgX = radY * radY;
    chgY = radX * radX * (1 - 2 * radY);;
    err = 0;
    stopX = 0;
    stopY = sqrX * radY;

    while (stopX <= stopY) {
      ++x;
      stopX += sqrY;
      err += chgX;
      chgX += sqrY;
      if (2 * err + chgY > 0) {
        drawCircLine(centerX, centerY, x - 1, oldX, y, pixel, part);
        oldX = x;
        --y;
        stopY -= sqrX;
        err += chgY;
        chgY += sqrX;
      }
    }
    
    // Lcken an den Sten zwischen oberer und unterer Schleife fllen,
    // vor allem bei extrem flachen Ellipsen
    for ( ; y > lastY; --y) {
      drawCircLine(centerX, centerY, x - 1, oldX, y, pixel, part);
    }
    if ((lastX > x) && !(part & SDL_CIRC_FILLED)) {
      drawCircLine(centerX, centerY, lastX - 1, x, lastY, pixel, part);
    }
  }

  if (SDL_MUSTLOCK(myScreen)) {
    SDL_UnlockSurface(myScreen);
  }
  
  // ev. modifizierten Bereich vergrern
  
  if (leftX < minX) minX = leftX;
  if (upperY < minY) minY = upperY;
  if (rightX > maxX) maxX = rightX;
  if (lowerY > maxY) maxY = lowerY;
}

static void sdlQuitOnEvent(void)
{
  SDL_Event event;
  while (SDL_PollEvent(&event)) {
    if (event.type == SDL_QUIT) {
      exit(EXIT_FAILURE);
    }
  }
}

void sdlUpdate(void)
{
  sdlQuitOnEvent();
  // Wenn seit letztem Mal nichts gendert wurde: Kein Update  
  if (maxX > MINVAL) {
    SDL_UpdateRect(myScreen, minX, minY,
                   (maxX - minX) + 1, (maxY - minY) + 1);
    minX = MAXVAL;
    maxX = MINVAL;
    minY = MAXVAL;
    maxY = MINVAL;
  }
}

void sdlMilliSleep(int ms)
{
  sdlQuitOnEvent();
  SDL_Delay(ms);
}
